# Login code adapted from the flaskr tutorial (see LICENSE-3RD-PARTY.md for details)
# See https://flask.palletsprojects.com/en/2.0.x/tutorial/views/
import functools
from typing import Any, Callable

from flask import g, request

from portal import db

from ..model import Key, User
from ..problem_details import ProblemResponse, not_found, unauthorized
from .blueprint import bp


@bp.before_app_request
def set_user_if_valid() -> None:
    if (
        not request.authorization
        or request.authorization.type != "basic"
        or request.authorization.username is None
        or request.authorization.password is None
    ):
        g.user = None
        return

    g.user = User.authenticate(
        request.authorization.username, request.authorization.password
    )


@bp.before_app_request
def set_key_if_valid() -> None:
    if (
        not request.authorization
        or request.authorization.type != "bearer"
        or request.authorization.token is None
    ):
        g.key = None
        return

    g.key = Key.get_by_secret_unless_expired(request.authorization.token)


def require_valid(credential_type: type[User | Key]) -> Callable:
    """
    Requires that valid credentials are provided before allowing access to a route.
    If no valid credentials are provided, abort and send a 401.

    This decorator can be used in two ways:
    - `@require_valid(User)` to require a valid username and password
    - `@require_valid(Key)` to require a valid key
    """

    def decorator(route: Callable) -> Callable:
        @functools.wraps(route)
        def protected_route(**kwargs: Any) -> Any:
            # Determine which credentials to use
            credentials = {User: g.user, Key: g.key}.get(credential_type)

            if credentials is None:
                return unauthorized()
            else:
                # If g.user or g.key is set, we can trust they are valid.
                return route(**kwargs)

        return protected_route

    return decorator


@bp.post("/keys")
@require_valid(User)
def post_key() -> tuple[dict, int]:
    key, secret = Key.generate(g.user)
    return key.to_dict() | {"secret": secret}, 201


@bp.delete("/keys/<string:uuid>")
@require_valid(Key)
def delete_key(uuid: str) -> tuple[str, int] | ProblemResponse:
    key_to_delete = Key.get_only(uuid)

    if not key_to_delete:
        return not_found()

    if key_to_delete.user_uuid != g.key.user_uuid:
        # If the key exists, but is from a different user, we pretend it isn't there
        return not_found()

    db.session.delete(key_to_delete)
    db.session.commit()

    return "", 204
